import { Message } from "revolt.js";
import { spawn } from 'child_process';
import { embed } from "./index.js";
import { SendableEmbed } from "revolt-api";
import FormData from 'form-data';
import axios from "axios";

let IGNORED = ["OK", "G2G"];
let RUNNING: string[] = [];

export default async function sherlock(message: Message, args: string[], nsfw: boolean) {
    const SHERLOCK_COMMAND = (process.env.SHERLOCK_COMMAND || 'python3 sherlock').split(' ');

    if (RUNNING.includes(message.channel_id)) {
        return await message.reply({ embeds: [
            embed('Please wait for the previous scan to finish before launching another one!', 'Sherlock', 'ERROR'),
        ] });
    }

    return new Promise<void>(async (resolve, reject) => {
        try {
            if (!args.length) {
                await message.reply({ embeds: [
                    embed('No username(s) to search for provided.', 'Sherlock', 'ERROR'),
                ] });
                return resolve();
            }

            if (args.length > 3) {
                await message.reply({ embeds: [
                    embed('Please provide no more than 3 usernames.', 'Sherlock', 'ERROR'),
                ] });
                return resolve();
            }

            RUNNING.push(message.channel_id);

            const infoSearchingEmbed = embed(
                `Looking up \`${args.join('`, `')}\`. This might take a while, please be patient!\n\nIncluding NSFW results: \`${nsfw}\``,
                'Sherlock',
                'INFO',
            );
            const msg = await message.reply({ embeds: [ infoSearchingEmbed ] });
            const matches: { user: string, name: string, url: string }[] = [];
            const searchedEmbeds: SendableEmbed[] = [];
            const attachments: string[] = [];
            let currentUser = '';

            const addEmbed = async (name: string) => {
                let text = `Found ${matches.filter(match => match.user == name).length} links!\n`;
                for (const match of matches.filter(match => match.user == name)) {
                    const newLine = args.length > 1
                        ? `\n- [${match.name}](${match.url.replaceAll(' ', '%20')})`
                        : `\n$\\textsf{\\color{#1abc9c}${match.name}}$: ${match.url.replaceAll(' ', '%20')}`;

                    if (text.length + newLine.length > 2000) break;
                    else if (text.length + newLine.length > 1800 / args.length) {
                        text = 'Message too long - Results will be sent as file';
                        attachments.push(await uploadFile(
                            matches.filter(match => match.user == name)
                                .map(m => `${m.name}: ${m.url}`)
                                .join('\n'),
                            name,
                        ));
                        break;
                    }
                    else text += newLine;
                }

                searchedEmbeds.push(embed(
                    text,
                    `Sherlock - ${name}`,
                    'SUCCESS'
                ));
            }

            const previousLength = matches.length;
            const interval = setInterval(async () => {
                try {
                    if (matches.length > previousLength) {
                        await msg?.edit({ embeds: [
                            infoSearchingEmbed,
                            ...searchedEmbeds,
                            embed(
                                `Searching: \`${currentUser}\` - ${matches.filter(m => m.user == currentUser).length} found so far`,
                                `Sherlock - ${currentUser}`,
                                'INFO',
                            ),
                        ] });
                    }
                } catch(e) {
                    // Would be super funny if the errors were actually fucking catched, wouldn't it?
                    clearInterval(interval);
                    proc.kill();
                    RUNNING = RUNNING.filter(r => r != message.channel_id);
                    reject(e);
                }
            }, 3000);

            const cmdArgs = nsfw
                ? [ ...SHERLOCK_COMMAND.slice(1), '--nsfw', '--output', '/dev/null', '--', ...args ]
                : [ ...SHERLOCK_COMMAND.slice(1), '--output', '/dev/null', '--', ...args ];
            console.log(`Invoking: ${SHERLOCK_COMMAND[0]} ${cmdArgs.join(' ')}`);

            const proc = spawn(`${SHERLOCK_COMMAND[0]}`, cmdArgs, { shell: false });

            proc.on('exit', async (res) => {
                console.log(`[Sherlock] Exited with code ${res}`);
                RUNNING = RUNNING.filter(r => r != message.channel_id);
                clearInterval(interval);

                if (res != 0) {
                    await msg?.edit({ embeds: [
                        embed(`Sherlock exited with code ${res}.\n\`\`\`\n${stderr}\n\`\`\``, 'Sherlock - Error', 'ERROR'),
                    ] });
                }
                else {
                    await addEmbed(currentUser);
                    await msg?.edit({ embeds: searchedEmbeds });
                    if (attachments.length) {
                        await message.channel?.sendMessage({ attachments });
                    }
                }
                return resolve();
            });

            let stderr = '';
            proc.stderr.on('data', (data) => {
                console.log('[Sherlock] [stderr] ' + data.toString('utf8'));
                stderr += data.toString('utf8');
            });

            let stdoutBuf = '';
            proc.stdout.on('data', async (data) => {
                stdoutBuf += data.toString('utf8');

                if (stdoutBuf.includes('\n')) {
                    const buf = stdoutBuf.split('\n');
                    stdoutBuf = buf.pop() ?? '';

                    for (let line of buf) {
                        console.log('[Sherlock] [stdout] ' + line);
                        if (line.startsWith('[+]')) {
                            line = line.substring(3).trimStart();
                            const [name, url] = line.split(': ');
                            if (!IGNORED.includes(name)) {
                                matches.push({ user: currentUser, name, url });
                            }
                        }
                        else if (line.startsWith('[*] Checking username')) {
                            line = line.substring('[*] Checking username'.length, line.length - 'on:'.length).trim();

                            if (currentUser) {
                                await addEmbed(currentUser);
                            }

                            currentUser = line;
                            console.log('[Sherlock] Now scanning user:', currentUser);
                        }
                    }
                }
            });

        } catch(e) {
            reject(e);
            RUNNING = RUNNING.filter(r => r != message.channel_id);
        }
    });
}

async function uploadFile(file: any, filename: string): Promise<string> {
    let data = new FormData();
    data.append("file", file, { filename: filename });

    let req = await axios.post("https://autumn.revolt.chat/attachments", data, {
        headers: data.getHeaders(),
    });
    return (req.data as any)["id"] as string;
}
